// LLVM includes
#include "llvm/ADT/DepthFirstIterator.h"
#include "llvm/ADT/PostOrderIterator.h"
#include "llvm/ADT/SCCIterator.h"
#include "llvm/ADT/StringRef.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/GlobalObject.h"
#include "llvm/IR/IRBuilder.h"
#include "llvm/IR/MDBuilder.h"
#include "llvm/IR/Module.h"
#include "llvm/IR/PassManager.h"
#include "llvm/Pass.h"
#include "llvm/Support/Error.h"
#include "llvm/Support/FileSystem.h"
#include "llvm/Transforms/Utils/Cloning.h"
#include "llvm/Transforms/Utils/CodeExtractor.h"

// Local includes
#include "LLVMHelpers.h"

using namespace llvm;

class UncommonCaseOutlinerPass
  : public llvm::PassInfoMixin<UncommonCaseOutlinerPass> {
public:
  llvm::PreservedAnalyses run(llvm::Module &M, llvm::ModuleAnalysisManager &);
};

// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
llvm::PreservedAnalyses
UncommonCaseOutlinerPass::run(Module &M, ModuleAnalysisManager &) {
  Function *CommonCaseFn = M.getFunction("common_case");

  Error Err = Error::success();

  ExitOnError ExitOnErr;
  ExitOnErr.setBanner("Error: Failed to perform uncommon case extraction. Make "
                      "sure that the common/uncommon case markers are not "
                      "inside loops or conditional branches. See the following "
                      "lines for more details.\n");

  // Find and collect all calls to the common_case() function. Also ensure that
  // there is only one call per function.

  // Work on each common case call / caller function
  for (Function &F : M) {
    SmallVector<CallInst *, 8> CommonCaseCalls;

    if (F.isDeclaration())
      continue;

    // Find all calls to common_case() in the function, in order of appearance
    for (BasicBlock *BB : post_order(&F.getEntryBlock())) {
      for (Instruction &I : *BB) {
        auto *Call = dyn_cast<CallInst>(&I);
        if (Call != nullptr && Call->getCalledFunction() == CommonCaseFn)
          CommonCaseCalls.push_back(Call);
      }
    }

    // Handle each common case call, in reverse order
    for (CallInst *CommonCaseCall : CommonCaseCalls) {
      // Validate that the common_case call is not inside a loop
      BasicBlock *CommonCaseCallBB = CommonCaseCall->getParent();
      bool InLoop = false;
      auto It = scc_begin(&F);
      for (; !It.isAtEnd() && !InLoop; ++It) {
        if (!It.hasCycle())
          continue;
        for (BasicBlock *BB : *It) {
          if (BB == CommonCaseCallBB) {
            InLoop = true;
            break;
          }
        }
      }
      if (InLoop) {
        Err = joinErrors(std::move(Err),
                         make_error<StringError>("common_case() or similar is "
                                                 "in "
                                                 "a loop in function '" +
                                                   F.getName() + "'",
                                                 inconvertibleErrorCode()));
        continue;
      }

      // Split the basic block after the common case call, moving all subsequent
      // instructions into a new basic block
      BasicBlock *FollowingBB = CommonCaseCall->getParent();
      FollowingBB->splitBasicBlockBefore(CommonCaseCall->getNextNode(),
                                         FollowingBB->getName() + ".check_"
                                                                  "cond");

      // Find all BBs that we want to move into the new function by collecting
      // FollowingBB and all its successors.
      SmallVector<BasicBlock *, 32> BBsToExtract;
      for (BasicBlock *BB : depth_first(FollowingBB)) {
        BBsToExtract.push_back(BB);
      }

      // Ensure that the BBs to be extracted don't contain any calls to the
      // common_case() function. They shouldn't, since we're iterating in
      // post order.
      for (BasicBlock *BB : BBsToExtract) {
        for (Instruction &I : *BB) {
          auto *Call = dyn_cast<CallInst>(&I);
          if (Call != nullptr && Call->getCalledFunction() == CommonCaseFn)
            abort();
        }
      }

      CodeExtractor CE(BBsToExtract);
      CodeExtractorAnalysisCache CEAC(F);

      // Error-out early if the CodeExtractor thinks it's not possible to
      // extract the code region
      if (!CE.isEligible()) {
        Err = joinErrors(std::move(Err),
                         make_error<StringError>("Could not isolate uncommon "
                                                 "case in '" +
                                                   F.getName() +
                                                   "', make sure the "
                                                   "common_case/fast_path/"
                                                   "hopefully call is not "
                                                   "inside an if or a loop",
                                                 inconvertibleErrorCode()));
        continue;
      }

      // Extract the code region
      SetVector<Value *> Inputs;
      SetVector<Value *> Outputs;
      Function *ExtractedCommonFn = CE.extractCodeRegion(CEAC, Inputs, Outputs);

      assert(ExtractedCommonFn != nullptr && "Could not extract function");
      assert(ExtractedCommonFn->getReturnType()->isVoidTy());

      // Find the call to the new extracted function in the original function
      // and the basic block that contains it, for later use
      BasicBlock *ExtractedFnTrampolineBB = nullptr;
      CallInst *ExtractedFnCall = nullptr;
      for (BasicBlock &BB : F) {
        for (Instruction &I : BB) {
          if (auto *CI = dyn_cast<CallInst>(&I)) {
            if (CI->getCalledFunction() == ExtractedCommonFn) {
              ExtractedFnTrampolineBB = &BB;
              ExtractedFnCall = CI;
              break;
            }
          }
        }
      }
      assert(ExtractedFnTrampolineBB != nullptr);
      assert(pred_size(ExtractedFnTrampolineBB) == 1);

      // Find which argument of the extracted function is the return value of
      // the common_case() call. That's the favorable branch marker condition.
      int ConditionArgIdx = -1;
      for (int i = 0; i < ExtractedFnCall->arg_size(); ++i) {
        if (ExtractedFnCall->getArgOperand(i) == CommonCaseCall) {
          ConditionArgIdx = i;
          break;
        }
      }
      if (ConditionArgIdx < 0) {
        Err = joinErrors(std::move(Err),
                         make_error<StringError>("Could not find argument for "
                                                 "common_case() in function "
                                                 "extracted from '" +
                                                   F.getName() + "'",
                                                 inconvertibleErrorCode()));
        continue;
      }

      // If the extracted function contains output argument pointers, convert
      // one of them to a return value.
      for (Value *Output : Outputs) {
        // The Outputs vector contains the values that are stored into the out
        // arg pointers within the extracted function. We want to find the
        // matching alloca in the caller.

        // We expect exactly one use of the output value: a store instruction
        if (!Output->hasOneUse())
          continue;
        auto *OutputStore = dyn_cast<StoreInst>(*Output->user_begin());
        if (OutputStore == nullptr)
          continue;

        // Find the argument index by looking for the stored pointer in the args
        Value *OutputInnerArg = OutputStore->getPointerOperand();
        int OutArgIdx = -1;
        for (int j = 0; j < ExtractedCommonFn->arg_size(); ++j) {
          if (ExtractedCommonFn->getArg(j) == OutputInnerArg) {
            OutArgIdx = j;
            break;
          }
        }
        assert(OutArgIdx >= 0);

        // Find the alloca in the caller by finding the corresponding argument
        // in the call site
        auto *OutAllocaInCaller = dyn_cast<AllocaInst>(
          ExtractedFnCall->getArgOperand(OutArgIdx));
        if (OutAllocaInCaller == nullptr ||
            !FunctionType::isValidReturnType(
              OutAllocaInCaller->getAllocatedType()))
          continue;

        // Create a new function type that returns the output argument and that
        // does not have the output argument
        FunctionType *OldFTy = ExtractedCommonFn->getFunctionType();
        SmallVector<Type *, 8> ArgTypes(OldFTy->param_begin(),
                                        OldFTy->param_end());
        ArgTypes.erase(ArgTypes.begin() + OutArgIdx);
        FunctionType *NewFTy = FunctionType::get(OutAllocaInCaller
                                                   ->getAllocatedType(),
                                                 ArgTypes,
                                                 OldFTy->isVarArg());

        // Create a new function with the new type
        Function *NewFn = Function::Create(NewFTy,
                                           ExtractedCommonFn->getLinkage(),
                                           ExtractedCommonFn->getName(),
                                           M);
        ExtractedCommonFn->setName(ExtractedCommonFn->getName() + ".ret_"
                                                                  "outarg");

        // Mark the old function as alwaysinline, so it is inlined into the
        // wrapper
        ExtractedCommonFn->removeFnAttr(Attribute::NoInline);
        ExtractedCommonFn->addFnAttr(Attribute::AlwaysInline);

        // Implement the new function by allocating the output argument on the
        // stack, calling the old function, loading the output argument and
        // returning it. AlwaysInliner -> Mem2Reg -> DCE will remove the
        // unnecessary stack allocation.
        BasicBlock *AllocaBB = BasicBlock::Create(NewFn->getContext(),
                                                  "alloca",
                                                  NewFn);
        BasicBlock *CallBB = BasicBlock::Create(NewFn->getContext(),
                                                "call",
                                                NewFn);
        IRBuilder<> Builder(AllocaBB);

        // The AllocaBB performs the allocation then branches to the CallBB
        AllocaInst *Alloca = Builder.CreateAlloca(
          OutAllocaInCaller->getAllocatedType());
        Builder.CreateBr(CallBB);
        Builder.SetInsertPoint(CallBB);

        // The CallBB calls the old function, loads the output argument and
        // returns it
        SmallVector<Value *, 8> ProxiedArgs;
        for (int Idx = 0; Idx < ExtractedCommonFn->arg_size(); ++Idx) {
          if (Idx == OutArgIdx)
            ProxiedArgs.push_back(Alloca);
          else
            ProxiedArgs.push_back(
              NewFn->getArg(Idx - (Idx > OutArgIdx ? 1 : 0)));
        }
        Builder.CreateCall(ExtractedCommonFn, ProxiedArgs);
        Value *OutVal = Builder.CreateLoad(Alloca->getAllocatedType(), Alloca);
        Builder.CreateRet(OutVal);

        // Fix the call site by creating a new call instruction that saves the
        // return value and does not pass the output argument.
        Builder.SetInsertPoint(ExtractedFnCall);
        SmallVector<Value *, 8> NewCallArgs;
        for (int Idx = 0; Idx < ExtractedFnCall->arg_size(); ++Idx) {
          if (Idx == OutArgIdx)
            continue;
          NewCallArgs.push_back(ExtractedFnCall->getArgOperand(Idx));
        }
        CallInst *NewCall = Builder.CreateCall(NewFn, NewCallArgs, "retval");

        // Store the return value into the output argument alloca created by the
        // CodeExtractor. Mem2Reg will remove the alloca and the store.
        Builder.CreateStore(NewCall, OutAllocaInCaller);

        ExtractedFnCall->eraseFromParent();

        // Use this new function as the extracted function
        ExtractedCommonFn = NewFn;
        ExtractedFnCall = NewCall;
      }

      // Create a copy of the extracted function for the uncommon case
      ValueToValueMapTy VMap;
      Function *ExtractedUncommonFn = CloneFunction(ExtractedCommonFn, VMap);

      // The common case should always be inlined
      ExtractedCommonFn->setName(F.getName() + ".common_case");
      ExtractedCommonFn->setLinkage(GlobalValue::InternalLinkage);
      ExtractedCommonFn->removeFnAttr(Attribute::NoInline);
      ExtractedCommonFn->addFnAttr(Attribute::AlwaysInline);

      // We mark the uncommon case as noinline; we will remove it later in the
      // late optimization passes.
      ExtractedUncommonFn->setName(F.getName() + ".uncommon_case");
      ExtractedUncommonFn->setLinkage(GlobalValue::InternalLinkage);
      ExtractedUncommonFn->removeFnAttr(Attribute::AlwaysInline);
      ExtractedUncommonFn->addFnAttr(Attribute::NoInline);

      // In the common and uncommon case functions, replace the condition
      // argument usages respectively with a constant true/false. We leave the
      // unused argument as it is, LLVM will remove it.
      ExtractedCommonFn->getArg(ConditionArgIdx)
        ->replaceAllUsesWith(
          ConstantInt::getTrue(ExtractedCommonFn->getContext()));
      ExtractedUncommonFn->getArg(ConditionArgIdx)
        ->replaceAllUsesWith(
          ConstantInt::getFalse(ExtractedUncommonFn->getContext()));

      // At the call site, we replace the condition argument so that instead of
      // passing the return value of common_case(), we pass its argument which
      // is the actual condition.
      Value *Condition = CommonCaseCall->getArgOperand(ConditionArgIdx);
      CommonCaseCall->replaceAllUsesWith(Condition);

      // Create always-inline proxy function that calls the common and uncommon
      // case functions depending on the condition. Using a proxy function
      // allows us to not take care of any phi nodes we might need to insert
      // into the original function. AlwaysInliner will take care of that for
      // us.
      Function *Proxy = Function::Create(ExtractedCommonFn->getFunctionType(),
                                         llvm::GlobalValue::InternalLinkage,
                                         F.getName() + ".proxy",
                                         &M);
      Proxy->removeFnAttr(Attribute::NoInline);
      Proxy->addFnAttr(Attribute::AlwaysInline);

      // The proxy function does this:
      // if (condition) {
      //   return common_case(args);
      // } else {
      //   return uncommon_case(args);
      // }
      SmallVector<Value *, 8> ProxyArgs;
      for (Argument &Arg : Proxy->args()) {
        ProxyArgs.push_back(&Arg);
      }
      BasicBlock *EntryBB = BasicBlock::Create(M.getContext(), "entry", Proxy);
      BasicBlock *CommonCaseBB = BasicBlock::Create(M.getContext(),
                                                    "common_case",
                                                    Proxy);
      BasicBlock *UncommonCaseBB = BasicBlock::Create(M.getContext(),
                                                      "uncommon_case",
                                                      Proxy);
      BasicBlock *ReturnBB = BasicBlock::Create(M.getContext(),
                                                "return",
                                                Proxy);
      IRBuilder<> Builder(EntryBB);
      Builder.CreateCondBr(Proxy->getArg(ConditionArgIdx),
                           CommonCaseBB,
                           UncommonCaseBB);
      Builder.SetInsertPoint(CommonCaseBB);
      CallInst *ComCall = Builder.CreateCall(ExtractedCommonFn, ProxyArgs);
      Builder.CreateBr(ReturnBB);
      Builder.SetInsertPoint(UncommonCaseBB);
      CallInst *UncomCall = Builder.CreateCall(ExtractedUncommonFn, ProxyArgs);
      Builder.CreateBr(ReturnBB);

      Builder.SetInsertPoint(ReturnBB);
      FunctionType *FTy = ExtractedCommonFn->getFunctionType();
      if (FTy->getReturnType()->isVoidTy()) {
        Builder.CreateRetVoid();
      } else {
        ComCall->setName("retval.common");
        UncomCall->setName("retval.uncommon");
        PHINode *Phi = Builder.CreatePHI(FTy->getReturnType(), 2, "retval");
        Phi->addIncoming(ComCall, CommonCaseBB);
        Phi->addIncoming(UncomCall, UncommonCaseBB);
        Builder.CreateRet(Phi);
      }

      // Replace the callee in the cloned basic block with the proxy function
      bool FoundCall = false;
      for (Instruction &I : *ExtractedFnTrampolineBB) {
        if (auto *CI = dyn_cast<CallInst>(&I)) {
          if (CI->getCalledFunction() == ExtractedCommonFn) {
            CI->setCalledFunction(Proxy);
            FoundCall = true;
            break;
          }
        }
      }

      // Remove the call to common_case() from the original function
      CommonCaseCall->eraseFromParent();

      assert(FoundCall);
    }
  }

  ExitOnErr(std::move(Err));

  return llvm::PreservedAnalyses::none();
}

struct UncommonCaseOutliner
  : public BasicLegacyPassWrapper<UncommonCaseOutlinerPass> {
  static char ID;
  UncommonCaseOutliner() :
    BasicLegacyPassWrapper<UncommonCaseOutlinerPass>(ID) {}
};

char UncommonCaseOutliner::ID = 0;
static RegisterPass<UncommonCaseOutliner> X("uncommon-case-outliner",
                                            "Outlines uncommon cases marked in "
                                            "the source code",
                                            false /* Only looks at CFG */,
                                            false /* Analysis Pass */);
